import os

import torch
import orjsonl
from transformers import CLIPConfig, CLIPModel, CLIPProcessor
from PIL import Image
from tqdm import trange, tqdm

from .base import MetricRunner, batch_load
from kgen.utils import remove_repeated_suffix


class CLIPMetricRunner(MetricRunner):
    def __init__(self, model_id="zer0int/LongCLIP-GmP-ViT-L-14", dtype=torch.float16):
        self.dtype = torch.float16
        self.model_id = "zer0int/LongCLIP-GmP-ViT-L-14"
        self.config = CLIPConfig.from_pretrained(model_id)
        self.config.text_config.max_position_embeddings = 248
        self.clip_model = (
            CLIPModel.from_pretrained(
                model_id,
                torch_dtype=dtype,
                config=self.config,
                device_map="cuda",
                attn_implementation="flash_attention_2",
            )
            .requires_grad_(False)
            .half()
        )
        self.clip_processor = CLIPProcessor.from_pretrained(
            model_id, padding="max_length", max_length=248
        )

    @torch.no_grad()
    def eval(self, images, ref_texts=None, is_ref=False):
        inputs = self.clip_processor(
            text=ref_texts,
            images=images,
            return_tensors="pt",
            truncation=True,
            padding=True,
        )
        for k, v in inputs.items():
            if isinstance(v, torch.Tensor):
                inputs[k] = v.cuda()
        outputs = self.clip_model(**inputs)
        cosine_sim = outputs.logits_per_image / self.clip_model.logit_scale.exp()
        return cosine_sim[torch.eye(cosine_sim.shape[0]).bool()]

    def eval_multi(self, images, ref_texts=None, ref_images=None, batch_size=32):
        results = super().eval_multi(images, ref_texts, ref_images, batch_size)
        results = torch.concat(results, dim=0)
        return torch.mean(results, dim=0)


if __name__ == "__main__":

    def load_prompts(file):
        datas = []
        for data in orjsonl.load(file):
            org_data = data["entry"]
            index = org_data["key"]
            result1 = data["result1"]
            result2 = data["result2"]
            org_prompt1 = remove_repeated_suffix(
                org_data["caption_llava_short"].strip()
            )
            org_prompt2 = ".".join(
                remove_repeated_suffix(org_data["caption_llava"].strip()).split(".")[:2]
            )
            gen_prompt1 = result1["generated"]
            gen_prompt2 = result2["extended"]
            datas.append((index, org_prompt1, org_prompt2, gen_prompt1, gen_prompt2))
        return datas

    def load_prompts_gbc(file):
        datas = []
        for data in orjsonl.load(file):
            org_data = data["entry"]
            index = org_data["index"]
            result1 = data["result1"]
            result2 = data["result2"]
            org_prompt1 = remove_repeated_suffix(org_data["short_caption"].strip())
            org_prompt2 = ".".join(
                remove_repeated_suffix(org_data["detail_caption"].strip()).split(".")[
                    :2
                ]
            )
            gen_prompt1 = result1["generated"]
            gen_prompt2 = result2["extended"]
            datas.append((index, org_prompt1, org_prompt2, gen_prompt1, gen_prompt2))
        return datas

    runner = CLIPMetricRunner()
    texts = [[], [], [], []]
    images = [[], [], [], []]

    datas = load_prompts_gbc("./data/gbc-output.jsonl")
    img_files = os.listdir("./data/gbc-img")
    for index, org_prompt1, org_prompt2, gen_prompt1, gen_prompt2 in tqdm(
        datas, total=len(datas), desc="load"
    ):
        org_img1 = f"./data/gbc-img/{index}_short.webp"
        org_img2 = f"./data/gbc-img/{index}_truncate_long.webp"
        gen_img1 = f"./data/gbc-img/{index}_tipo_gen.webp"
        gen_img2 = f"./data/gbc-img/{index}_tipo_extend.webp"
        texts[0].append(org_prompt1)
        texts[1].append(org_prompt2)
        texts[2].append(gen_prompt1)
        texts[3].append(gen_prompt2)
        images[0].append(org_img1)
        images[1].append(org_img2)
        images[2].append(gen_img1)
        images[3].append(gen_img2)

    for idx, name in enumerate(["short", "truncate", "tipo gen", "tipo extend"]):
        result = runner.eval_multi(images[idx], texts[idx], batch_size=500)
        print(idx, name, result.shape, torch.mean(result))



import os
import re
import pathlib

from .metainfo import SPECIAL, POSSIBLE_QUALITY_TAGS, RATING_TAGS


tag_list_folder = pathlib.Path(os.path.dirname(__file__)) / "tag-list"
tag_lists = {
    os.path.splitext(f)[0]: set(open(tag_list_folder / f).read().strip().split("\n"))
    for f in os.listdir(tag_list_folder)
    if f.endswith(".txt")
}
tag_lists["special"] = set(SPECIAL)
tag_lists["quality"] = set(POSSIBLE_QUALITY_TAGS)
tag_lists["rating"] = set(RATING_TAGS)


def seperate_tags(all_tags):
    all_tags = [i.strip() for i in all_tags if i.strip()]
    tag_map = {cate: [] for cate in tag_lists.keys()}
    tag_map["general"] = []
    for tag in all_tags:
        for cate in tag_lists.keys():
            if tag in tag_lists[cate]:
                tag_map[cate].append(tag)
                break
            if tag.replace("_", " ") in tag_lists[cate]:
                tag_map[cate].append(tag.replace("_", " "))
                break
        else:
            if len(tag) < 4:
                tag_map["general"].append(tag)
            else:
                tag_map["general"].append(tag.replace("_", " "))
    return tag_map


redundant_form: list[tuple[re.Pattern, str]] = [
    [re.compile(r"\n +"), "\n"],
    [re.compile(r"\n\n\n+"), "\n\n"],
    [re.compile(r"  +"), " "],
]


def apply_format(tag_map, form):
    if "<|extended|>" in form and not tag_map.get("extended", ""):
        form = form.replace("<|extended|>", "<|generated|>")
    for type in tag_map:
        if f"<|{type}|>" in form:
            if not tag_map[type]:
                form = form.replace(f"<|{type}|>,", "")
                form = form.replace(f"<|{type}|>", "")
            else:
                data = tag_map[type] or ""
                if isinstance(data, list):
                    form = form.replace(f"<|{type}|>", ", ".join(data))
                else:
                    form = form.replace(f"<|{type}|>", data)
    form = re.sub(r"<\|.*\|>", "", form)
    for pattern, repl in redundant_form:
        form = pattern.sub(repl, form)
    return form.strip().strip(",")



import os
import torch
import orjsonl
import orjson

from aesthetic_predictor_v2_5 import convert_v2_5_from_siglip
from .base import MetricRunner, load


class AestheticRunner(MetricRunner):
    single = True
    multi = True

    def __init__(self):
        # load model and preprocessor
        model, preprocessor = convert_v2_5_from_siglip(
            low_cpu_mem_usage=False,
            trust_remote_code=True,
            attn_implementation="flash_attention_2",
            torch_dtype=torch.float32,
        )
        model = model.cuda().requires_grad_(False)
        self.model = model
        self.preprocessor = preprocessor

    def img_load_func(self, image):
        image = load(image)
        pixel_values = self.preprocessor(images=image, return_tensors="pt").pixel_values
        return pixel_values

    @torch.inference_mode()
    def eval(self, images, ref_texts=None, is_ref=False):
        pixel_values = torch.concat(images).cuda().float()
        with torch.autocast("cuda", dtype=torch.float16):
            result = self.model(pixel_values).logits.squeeze().float().cpu()
        assert torch.all(~torch.isnan(result)), f"nan in result: {torch.sum(torch.isnan(result))}/{result.numel()}"
        return result

    @torch.no_grad()
    def eval_multi(self, images, ref_texts=None, ref_images=None, batch_size=32):
        results, ref_results = super().eval_multi(images, ref_texts, ref_images, batch_size)
        results = torch.concat(results, dim=0).reshape(-1)
        return [i.item() for i in results]


if __name__ == "__main__":
    runner = AestheticRunner()
    images = [None] * 4

    PATH = "./data/coyo-img"
    img_files = [os.path.join(PATH, i) for i in os.listdir(PATH)]
    images[0] = [i for i in img_files if i.endswith("_short.webp")]
    images[1] = [i for i in img_files if i.endswith("_truncate_long.webp")]
    images[2] = [i for i in img_files if i.endswith("_gen.webp")]
    images[3] = [i for i in img_files if i.endswith("_extend.webp")]

    with open("gbc-aesthetic-v25.jsonl", "w") as f:
        for idx, name in enumerate(["short", "truncate", "tipo gen", "tipo extend"]):
            result = runner.eval_multi(images[idx], batch_size=350)
            f.write(str(result) + "\n")



import re
import random

from .. import models
from ..generate import generate
from ..formatter import seperate_tags
from ..utils import shuffle_iterable, same_order_deduplicate
from ..metainfo import (
    TARGET_TIPO,
    TARGET_TIPO_MAX,
    TARGET_TIPO_NL,
    TARGET_TIPO_NL_MAX,
)


BAN_TAGS = []


def apply_tipo_prompt(meta, general, nl_prompt, mode, length, expand, gen_meta=False):
    content = {
        "tag": general,
    }
    if nl_prompt and (mode is None or "short" not in mode):
        content["long"] = nl_prompt
    elif nl_prompt:
        content["short"] = nl_prompt

    prompt = ""
    target = ""
    for key, value in meta.items():
        if value:
            prompt += f"{key}: {value}\n"

    if length:
        target += f"<|{length}|>"
    if mode:
        content_order = mode.split("_to_")
        target += f" <|{mode}|>"
    else:
        content_order = ["tag"]
    if gen_meta:
        target += " <|gen_meta|>"
    target = target.strip()
    prompt += f"target: {target}\n"

    for idx, key in enumerate(content_order):
        if key in content and content[key].strip():
            prompt += f"{key}: {content[key]}\n"
        elif idx < len(content_order) - 1:
            prompt += f"{key}: \n"
    if expand:
        prompt = prompt.strip()
    else:
        prompt += f"{key}:"
    return prompt


parse = re.compile(r"\n([^:\n]+):(.*(?:\n(?![^:\n]+:).*)*)")
TYPE_MAP = {
    "short": "extended",
    "long": "generated",
}


def parse_tipo_result(result: str):
    result = "\n" + result.strip("<s>").strip("</s>").strip()
    result_dict = {}
    for type, content in parse.findall(result):
        type = type.strip()
        content = content.strip()
        if TYPE_MAP.get(type, type) in result_dict:
            continue
        if type == "tag":
            tags = [i.strip() for i in content.split(",") if i.strip()]
            content = seperate_tags(tags)
            for key, value in content.items():
                value = [i for i in value if i]
                if key in result_dict:
                    if not isinstance(result_dict[key], list):
                        result_dict[key] = [result_dict[key]]
                    result_dict[key].extend(value)
                elif not value:
                    result_dict[key] = []
                else:
                    result_dict[key] = value
            result_dict["tag"] = tags
        elif type not in {"short", "long"}:
            result_dict[TYPE_MAP.get(type, type)] = [
                i.strip() for i in content.split(",") if i.strip()
            ]
        else:
            if content == "<|empty|>" or not content:
                continue
            result_dict[TYPE_MAP.get(type, type)] = content
    return result_dict


def parse_tipo_request(
    tag_map,
    nl_prompt="",
    expand_tags=True,
    expand_prompt=True,
    generate_extra_nl_prompt=True,
    tag_first=True,
    tag_length_target="",
    nl_length_target="",
    add_quality=True,
):
    rating = ", ".join(tag_map.get("rating", []))
    artist = ", ".join(tag_map.get("artist", []))
    characters = ", ".join(tag_map.get("characters", []))
    copyrights = ", ".join(tag_map.get("copyrights", []))
    general = ", ".join(tag_map.get("special", []) + tag_map.get("general", []))
    general = general.strip().strip(",")
    meta = {
        "meta": ", ".join(tag_map.get("meta", [])),
        "rating": rating or None,
        "artist": artist.strip() or None,
        "characters": characters.strip() or None,
        "copyrights": copyrights.strip() or None,
    }
    if (
        models.model_have_quality_info.get(models.current_model_name, None)
        or add_quality
    ):
        quality = ", ".join(tag_map.get("quality", []))
        meta["quality"] = quality
    tag_length = tag_length_target or "long"
    nl_length = nl_length_target or "long"

    # list of [mode, target_output, output_name, length_target, expand]
    # mode with None means tag only
    operations = []
    op_for_nl_gen = None
    match (general.strip(), nl_prompt.strip(), bool(expand_tags), bool(expand_prompt)):
        case ("", "", _, _):  # no input
            # no input, tag_only -> tag_to_long
            operations = [
                [None, tag_length, True],
            ]
            op_for_nl_gen = ["tag_to_long", nl_length, True]
        case (_, "", expand, _):  # tag only
            if expand:
                operations = [
                    [None, tag_length, True],
                ]
            op_for_nl_gen = ["tag_to_long", nl_length, False]
        case ("", _, _, expand):  # prompt only
            # long_to_tag -> short_to_tag_to_long
            # expand here means "expand 'long'"
            operations = [
                ["long_to_tag", tag_length, expand],
            ]
            op_for_nl_gen = ["short_to_tag_to_long", nl_length, False]
        case (_, _, False, False):
            op_for_nl_gen = ["short_to_tag_to_long", nl_length, False]
        case (_, _, True, False):
            operations = [
                ["short_to_tag", tag_length, True],
                ["short_to_tag_to_long", nl_length, False],
            ]
            op_for_nl_gen = ["short_to_tag_to_long", nl_length, False]
        case (_, _, False, True):
            operations = [
                ["tag_to_long", tag_length, True],
                ["tag_to_short_to_long", nl_length, False],
            ]
            op_for_nl_gen = ["tag_to_short_to_long", nl_length, False]
        case (_, _, True, True):
            if tag_first:
                operations = [
                    ["short_to_tag", tag_length, True],
                    ["tag_to_long", nl_length, True],
                ]
                op_for_nl_gen = ["short_to_tag_to_long", nl_length, False]
            else:
                operations = [
                    ["tag_to_long", nl_length, True],
                    ["long_to_tag", tag_length, True],
                ]
                op_for_nl_gen = ["tag_to_short_to_long", nl_length, False]
    if generate_extra_nl_prompt:
        operations.append(op_for_nl_gen)
    return meta, operations, general, nl_prompt


def tag_filter(tag):
    if any(b in tag or re.match(b, tag) is not None for b in BAN_TAGS):
        return False
    return True


def post_generate_process(parsed, meta, general, nl_prompt, mode, length, expand):
    if mode is None:
        parsed.pop("extended", None)
        parsed.pop("generated", None)
    else:
        if not "long" in mode:
            parsed.pop("generated", None)
        if not "short" in mode:
            parsed.pop("extended", None)
    if "generated" in parsed and nl_prompt and not parsed.get("extended", "").strip():
        parsed["extended"] = parsed.pop("generated")
    input_tags = [tag.strip() for tag in general.split(",")]
    input_prompts = [tag.strip() for tag in nl_prompt.split(".") if tag.strip()]

    input_generals = [tag for tag in parsed.get("general", []) if tag in input_tags]
    output_generals = shuffle_iterable(
        [
            tag
            for tag in parsed.get("general", [])
            if tag_filter(tag) and tag not in input_tags
        ]
    )
    output_nl_prompts = [
        tag.strip()
        for tag in parsed.get("extended", "").split(".")
        if tag_filter(tag.strip()) and tag.strip() not in input_prompts and tag.strip()
    ]
    if output_nl_prompts and input_prompts[-1] in output_nl_prompts[0]:
        input_prompts[-1] = output_nl_prompts.pop(0)
    if output_nl_prompts:
        output_nl_prompts = shuffle_iterable(output_nl_prompts[:-1]) + [
            output_nl_prompts[-1]
        ]

    max_length_tags = TARGET_TIPO_MAX[length]
    max_length_nl = TARGET_TIPO_NL_MAX[length]

    if len(input_generals) + len(output_generals) > max_length_tags:
        output_generals = output_generals[: max(max_length_nl - len(input_generals), 0)]
    if len(input_prompts) + len(output_nl_prompts) > max_length_nl:
        output_nl_prompts = output_nl_prompts[
            : max(max_length_nl - len(input_prompts), 0)
        ]
    if "generated" in parsed:
        generated_nl = [
            tag.strip()
            for tag in parsed.get("generated", "").split(".")
            if tag_filter(tag.strip())
            and tag.strip() not in input_prompts
            and tag.strip()
        ]
        if len(generated_nl) > max_length_nl:
            generated_nl = generated_nl[:max_length_nl]
    else:
        generated_nl = []

    new_general = input_generals + output_generals
    new_nl_prompt = input_prompts + output_nl_prompts

    parsed["general"] = same_order_deduplicate(new_general)
    parsed["extended"] = ". ".join(same_order_deduplicate(new_nl_prompt))
    if generated_nl:
        parsed["generated"] = ". ".join(same_order_deduplicate(generated_nl))

    return parsed


def retry_criteria(parsed, check_slice=slice(0, -1), length="long"):
    checks = [
        len(parsed.get("special", []) + parsed.get("general", [])),
        len(parsed.get("extended", "").split(".")),
        len(parsed.get("generated", "").split(".")),
    ]
    low_thresholds = [
        TARGET_TIPO[length],
        TARGET_TIPO_NL[length],
        TARGET_TIPO_NL[length],
    ]
    high_thresholds = [1000, 1000, 1000]

    result = all(
        l <= i <= h
        for l, i, h in list(zip(low_thresholds, checks, high_thresholds))[check_slice]
    )

    return result


def generate_with_retry(
    meta,
    general,
    nl_prompt,
    mode,
    length,
    expand,
    gen_meta,
    seed=0,
    max_retry=10,
    max_same_output=5,
    retry_criteria=retry_criteria,
    total_timing=None,
    get_timing_detail=True,
    **kwargs,
):
    iter_count = 0
    prev_output = set()
    same_output_count = 0
    while iter_count <= max_retry and same_output_count < max_same_output:
        if mode is not None:
            target = mode.split("_to_")[-1]
        else:
            target = "tag"
        prompt = apply_tipo_prompt(
            meta, general, nl_prompt, mode, length, expand, gen_meta
        )
        generation_setting = {
            "temperature": 0.35,
            "min_p": 0.1,
            "top_p": 0.95,
            "top_k": 60,
            "max_new_tokens": 512,
            "seed": seed + iter_count,
        }
        generation_setting.update(kwargs)
        result, input_token_count, token_generated = generate(
            prompt=prompt, **generation_setting
        )
        timing = {}
        timing["generate_pass"] = 1
        timing["generated_tokens"] = token_generated
        timing["input_tokens"] = input_token_count
        if get_timing_detail and hasattr(models.text_model, "export_time"):
            timing.update(models.text_model.export_time())
        if total_timing is not None:
            for key in timing:
                total_timing[key] = total_timing.get(key, 0) + timing[key]
        parsed = parse_tipo_result(result)
        parsed = post_generate_process(
            parsed, meta, general, nl_prompt, mode, length, expand
        )
        yield result, parsed
        if target == "long" and "generated" not in parsed:
            target = "short"

        slices_map = {
            "tag": slice(0, 1),
            "short": slice(1, 2),
            "long": slice(2, 3),
        }
        # print(mode, end=" ")
        if retry_criteria(parsed, slices_map.get(target, slice(0, -1)), length):
            break
        iter_count += 1
        if result in prev_output:
            same_output_count += 1
        else:
            same_output_count = 0
            prev_output.add(result)

        nl_prompt = (
            parsed.get("extended", "") or parsed.get("generated", "") or nl_prompt
        )
        general = ", ".join(parsed.get("special", []) + parsed.get("general", []))
        nl_prompt = nl_prompt.strip()
    yield result, parsed


def tipo_runner_generator(
    meta, operations, general, nl_prompt, gen_meta=False, **kwargs
):
    total_timing = {}
    for idx, (mode, length, expand) in enumerate(operations):
        is_last = idx == len(operations) - 1
        prompt = apply_tipo_prompt(
            meta, general, nl_prompt, mode, length, expand, gen_meta and is_last
        )
        if length is None and not expand:
            parsed = parse_tipo_result(prompt)
            break
        for result, parsed in generate_with_retry(
            meta,
            general,
            nl_prompt,
            mode,
            length,
            expand,
            gen_meta and is_last,
            seed=kwargs.pop("seed", 0) or random.randint(0, 2**32),
            total_timing=total_timing,
            **kwargs,
        ):
            yield parsed, total_timing
        if not is_last:
            if "generated" in parsed and nl_prompt:
                parsed["extended"] = parsed.pop("generated")
            nl_prompt = (
                parsed.get("generated", []) or parsed.get("extended", []) or nl_prompt
            )
            general = ", ".join(parsed.get("special", []) + parsed.get("general", []))
            nl_prompt = nl_prompt.strip()

    yield parsed, total_timing


def tipo_runner(meta, operations, general, nl_prompt, gen_meta=False, **kwargs):
    for parsed, timing in tipo_runner_generator(
        meta, operations, general, nl_prompt, gen_meta, **kwargs
    ):
        pass
    return parsed, timing



import math
from PIL import Image, ImageFont, ImageDraw


def create_image_grid(images, rows=None, cols=None):
    """
    Create a grid of images from a list of PIL Image objects.

    :param images: List of PIL Image objects
    :param rows: Number of rows in the grid (optional)
    :param cols: Number of columns in the grid (optional)
    :return: A new PIL Image object with the grid
    """
    # Determine the number of rows and columns
    n = len(images)
    if rows is None and cols is None:
        cols = math.ceil(math.sqrt(n))
        rows = math.ceil(n / cols)
    elif rows is None:
        rows = math.ceil(n / cols)
    elif cols is None:
        cols = math.ceil(n / rows)

    # Get the size of the first image
    w, h = images[0].size

    # Create a new image with the appropriate size
    grid = Image.new("RGB", size=(cols * w, rows * h))

    # Paste the images into the grid
    for i, img in enumerate(images):
        box = (i % cols * w, i // cols * h)
        grid.paste(img, box)

    return grid


def create_image_grid_with_prompts(
    image_prompt_pairs,
    rows=None,
    cols=None,
    font_path=None,
    font_size=20,
    prompt_height=30,
):
    """
    Create a grid of images with prompts from a list of (PIL Image, prompt) pairs.

    :param image_prompt_pairs: List of (PIL Image, prompt) tuples
    :param rows: Number of rows in the grid (optional)
    :param cols: Number of columns in the grid (optional)
    :param font_path: Path to a TTF font file (optional)
    :param font_size: Font size for the prompts (default: 20)
    :param prompt_height: Height of the area for each prompt (default: 30)
    :return: A new PIL Image object with the grid
    """
    # Determine the number of rows and columns
    n = len(image_prompt_pairs)
    if rows is None and cols is None:
        cols = math.ceil(math.sqrt(n))
        rows = math.ceil(n / cols)
    elif rows is None:
        rows = math.ceil(n / cols)
    elif cols is None:
        cols = math.ceil(n / rows)

    # Get the size of the first image
    w, h = image_prompt_pairs[0][0].size

    # Create a new image with the appropriate size
    grid = Image.new("RGB", size=(cols * w, rows * (h + prompt_height)), color="white")
    draw = ImageDraw.Draw(grid)

    # Load font
    if font_path:
        font = ImageFont.truetype(font_path, font_size)
    else:
        font = ImageFont.load_default()

    # Paste the images and draw prompts
    for i, (img, prompt) in enumerate(image_prompt_pairs):
        x = i % cols * w
        y = i // cols * (h + prompt_height)

        # Paste the image
        grid.paste(img, (x, y + prompt_height))

        # Draw the prompt
        prompt = prompt[: int(w / 7)]  # Truncate prompt if too long (rough estimate)
        text_width = draw.textlength(prompt, font=font)
        text_x = x + (w - text_width) / 2  # Center text
        draw.text((text_x, y), prompt, fill="black", font=font)

    return grid



import os
import pathlib

import torch
from huggingface_hub import hf_hub_download
from transformers import LlamaForCausalLM, LlamaTokenizer

from .logging import logger
from .patch import patch


text_model = None
tokenizer = None
current_model_name = None


model_dir = pathlib.Path(__file__).parent / "models"
model_list = [
    "KBlueLeaf/DanTagGen-delta-rev2",
    "KBlueLeaf/DanTagGen-delta",
    "KBlueLeaf/DanTagGen-beta",
    "KBlueLeaf/DanTagGen-alpha",
    "KBlueLeaf/DanTagGen-gamma",
]
gguf_name = [
    "ggml-model-Q6_K.gguf",
    "ggml-model-Q8_0.gguf",
    "ggml-model-f16.gguf",
]
model_have_quality_info = {
    "DanTagGen-delta-rev2": True,
    "DanTagGen-delta": True,
    "DanTagGen-beta": False,
    "DanTagGen-alpha": False,
    "DanTagGen-gamma": False,
}

tipo_model_list = [
    ("KBlueLeaf/TIPO-200M-ft", ["TIPO-200M-ft-F16.gguf"]),
    ("KBlueLeaf/TIPO-500M", ["TIPO-500M_epoch5-F16.gguf"]),
    ("KBlueLeaf/TIPO-200M", ["TIPO-200M-40Btok-F16.gguf"]),
]


try:
    from llama_cpp import Llama
except Exception:
    logger.warning("Llama-cpp-python cannot be imported")
    Llama = None


if not os.path.isdir(model_dir):
    os.makedirs(model_dir, exist_ok=True)


def download_gguf(model_name=model_list[0], gguf_name=gguf_name[-1]):
    logger.info(f"Downloading gguf model from {model_name}")
    result = hf_hub_download(
        model_name,
        gguf_name,
        repo_type="model",
        cache_dir=None,
        local_dir=model_dir,
        local_dir_use_symlinks=False,
    )
    new_name = model_dir / f"{model_name.split('/')[-1]}_{gguf_name}"
    if os.path.isfile(new_name):
        os.remove(new_name)
    os.rename(result, new_name)
    logger.info(f"Downloaded gguf model to {new_name}")
    return new_name


def list_gguf():
    files = [
        str(model_dir / file)
        for file in os.listdir(model_dir)
        if file.endswith(".gguf")
    ]
    return files


def load_model(
    model_name=model_list[0],
    gguf=False,
    device="cpu",
    subfolder=None,
    tokenizer_name=None,
    strict=True,
    **kwargs,
):
    global text_model, tokenizer, current_model_name
    if gguf:
        model_name = os.path.basename(model_name)
        model_repo_name = model_name.split("_")[0]
        current_model_name = model_repo_name
        try:
            assert Llama is not None
            text_model = Llama(
                str(model_dir / model_name),
                n_ctx=1024,
                n_gpu_layers=0 if device == "cpu" else 1000,
                verbose=False,
                **kwargs,
            )
            if tokenizer_name is not None:
                tokenizer = LlamaTokenizer.from_pretrained(tokenizer_name)
            else:
                tokenizer = None
            logger.info(f"Llama-cpp-python/gguf model {model_name} loaded")
            if device == "cuda":
                logger.warning(
                    "llama.cpp have reproducibility issue on cuda "
                    "(https://github.com/ggerganov/llama.cpp/pull/1346) "
                    "It is suggested to use cpu or "
                    "compile llama-cpp-python by yourself "
                    "and set GGML_CUDA_MAX_STREAMS in the file ggml-cuda.cu to 1."
                )
            return
        except Exception as e:
            if strict:
                raise e
            logger.warning(f"Llama-cpp-python/gguf model {model_name} load failed")
            model_name = f"KBlueLeaf/{model_repo_name}"
    logger.info(f"Using transformers model {model_name}")
    if device == "cuda" and not torch.cuda.is_available():
        device = "cpu"
    if subfolder is not None:
        kwargs["subfolder"] = subfolder
    text_model = patch(
        LlamaForCausalLM.from_pretrained(
            model_name,
            torch_dtype=torch.float16,
            attn_implementation="sdpa",
            **kwargs,
        )
        .requires_grad_(False)
        .eval()
        .to(device)
    )
    tokenizer = LlamaTokenizer.from_pretrained(
        tokenizer_name or model_name, subfolder=subfolder
    )
    current_model_name = model_name.split("/")[-1]
    logger.info(f"Model {model_name} loaded")


if __name__ == "__main__":
    model_file = download_gguf()
    load_model(model_file, gguf=True)



from functools import partial
from time import time_ns
from transformers import PreTrainedModel


def generate(self: PreTrainedModel, *args, **kwargs):
    self.t0 = time_ns()
    self.t_forward = []
    result = self.org_generate(*args, **kwargs)
    self.t1 = time_ns()
    return result


def forward(self: PreTrainedModel, *args, **kwargs):
    t0 = time_ns()
    result = self.org_forward(*args, **kwargs)
    t1 = time_ns()
    self.t_forward.append((t0, t1))
    return result


def export_time(self: PreTrainedModel):
    prompt_process = self.t_forward[0][1] - self.t_forward[0][0]

    total_eval = 0
    total_sampling = 0
    prev_t1 = self.t_forward[0][1]
    for t0, t1 in self.t_forward[1:]:
        total_eval += t1 - t0
        total_sampling += t0 - prev_t1
        prev_t1 = t1
    total_eval += self.t1 - self.t_forward[-1][1]

    timings = {
        "total": (self.t1 - self.t0) / 1e6,
        "prompt_process": prompt_process / 1e6,
        "total_eval": total_eval / 1e6,
        "total_sampling": total_sampling / 1e6,
    }

    return timings


def patch(model: PreTrainedModel):
    model.org_forward = model.forward
    model.org_generate = model.generate
    model.forward = partial(forward, model)
    model.generate = partial(generate, model)
    model.export_time = partial(export_time, model)
    return model



import torch
from transformers import (
    GenerationConfig,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    set_seed,
)
from . import models

try:
    from llama_cpp import Llama
except ImportError:

    class Llama:
        pass


def generate(
    model: PreTrainedModel | Llama = None,
    tokenizer: PreTrainedTokenizerBase = None,
    prompt="",
    temperature=0.5,
    top_p=0.95,
    top_k=45,
    repetition_penalty=1.17,
    max_new_tokens=128,
    autocast_gen=lambda: torch.autocast("cpu", enabled=False),
    **kwargs,
):
    if model is None:
        model = models.text_model
    if tokenizer is None:
        tokenizer = models.tokenizer
    if isinstance(model, Llama):
        # print(kwargs)
        result = model.create_completion(
            prompt,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            max_tokens=max_new_tokens,
            repeat_penalty=repetition_penalty or 1,
            seed=kwargs.get("seed", None),
        )
        prompt_tokens = result["usage"]["prompt_tokens"]
        completion_tokens = result["usage"]["completion_tokens"]
        return prompt + result["choices"][0]["text"], prompt_tokens, completion_tokens
    if "seed" in kwargs:
        set_seed(kwargs["seed"] % 2**32)

    torch.cuda.empty_cache()
    inputs = tokenizer(prompt, return_tensors="pt")
    input_ids = inputs["input_ids"].to(next(model.parameters()).device)
    generation_config = GenerationConfig(
        temperature=temperature,
        top_p=top_p,
        top_k=top_k,
        repetition_penalty=repetition_penalty,
        do_sample=True,
        min_new_tokens=2,
        **kwargs,
    )
    with torch.no_grad(), autocast_gen():
        generation_output = model.generate(
            input_ids=input_ids,
            generation_config=generation_config,
            max_new_tokens=max_new_tokens,
        )
    s = generation_output[0]
    output = tokenizer.decode(s)
    prompt_tokens = len(input_ids[0])
    completion_tokens = len(s) - prompt_tokens

    torch.cuda.empty_cache()
    return output, prompt_tokens, completion_tokens



import re
from random import shuffle
from contextlib import nullcontext

from transformers import set_seed

from .. import models
from ..utils import same_order_deduplicate
from ..generate import generate


def apply_dtg_prompt(tag_map, target="", aspect_ratio=1.0):
    special_tags = ", ".join(tag_map.get("special", []))
    rating = ", ".join(tag_map.get("rating", []))
    artist = ", ".join(tag_map.get("artist", []))
    characters = ", ".join(tag_map.get("characters", []))
    copyrights = ", ".join(tag_map.get("copyrights", []))
    general = ", ".join(tag_map.get("general", []))
    prompt = f"""
rating: {rating or '<|empty|>'}
artist: {artist.strip() or '<|empty|>'}
characters: {characters.strip() or '<|empty|>'}
copyrights: {copyrights.strip() or '<|empty|>'}
aspect ratio: {f"{aspect_ratio:.1f}" or '<|empty|>'}
target: {'<|' + target + '|>' if target else '<|long|>'}
general: {special_tags}, {general.strip().strip(",")}<|input_end|>
""".strip()

    if models.model_have_quality_info.get(models.current_model_name, None):
        quality = ", ".join(tag_map.get("quality", ["masterpiece"]))
        prompt = f"quality: {quality}\n{prompt}"

    return prompt


def black_list_match(tag, black_list):
    for b in black_list:
        if re.match(b, tag):
            return True
    return False


def tag_gen(
    text_model,
    tokenizer,
    prompt,
    prompt_tags,
    len_target,
    black_list,
    temperature=0.5,
    top_p=0.95,
    top_k=100,
    max_new_tokens=256,
    max_retry=20,
    max_same_output=15,
    seed=None,
):
    retry = max_retry
    llm_gen = ""

    set_seed(seed)

    iter_count = 0
    prev_output = set()
    same_output_count = 0
    while retry >= 0 and same_output_count < max_same_output:
        llm_gen = generate(
            model=text_model,
            tokenizer=tokenizer,
            prompt=prompt,
            temperature=temperature,
            top_p=top_p,
            top_k=top_k,
            repetition_penalty=None,
            max_new_tokens=max_new_tokens,
            stream_output=False,
            autocast_gen=nullcontext,
            prompt_lookup_num_tokens=10,
            pad_token_id=getattr(tokenizer, "eos_token_id", None),
            eos_token_id=getattr(tokenizer, "eos_token_id", None),
            seed=seed + iter_count,
        )[0]
        iter_count += 1
        llm_gen = llm_gen.replace("</s>", "").replace("<s>", "")
        orig_prompt = llm_gen.split("<|input_end|>")[0]
        extra = llm_gen.split("<|input_end|>")[-1].strip().strip(",")
        extra_tokens = [
            tok.strip()
            for tok in extra.split(",")
            if not black_list_match(tok.strip(), black_list)
        ]
        extra_tokens = same_order_deduplicate(extra_tokens)
        llm_gen = llm_gen.replace(extra, ", ".join(extra_tokens))

        yield llm_gen, extra_tokens, iter_count

        if set(extra_tokens) == prev_output:
            same_output_count += 1
            retry += 1
        else:
            same_output_count = 0
            prev_output = set(extra_tokens)

        if len(prompt_tags) + len(extra_tokens) < len_target:
            retry -= 1
            shuffle(extra_tokens)
            llm_gen = f"{orig_prompt}<|input_end|>{', '.join(extra_tokens)}"
            prompt = llm_gen.strip().replace("  <|", " <|")
        else:
            break
    yield llm_gen, extra_tokens, iter_count



import os
import sys
from concurrent.futures import ProcessPoolExecutor
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.transforms import transforms
from torchvision.transforms.functional import to_tensor
from torchmetrics.image import FrechetInceptionDistance
from PIL import Image
from tqdm import trange, tqdm

from .base import MetricRunner


def load_img(file, size):
    with Image.open(file) as img:
        return to_tensor(img.convert("RGB").resize(size, Image.BICUBIC))


class DINOv2FeatureExtractor(nn.Module):
    def __init__(self, name="vitl14", device="cuda"):
        super().__init__()
        self.model = (
            torch.hub.load("facebookresearch/dinov2", "dinov2_" + name)
            .to(device)
            .eval()
            .requires_grad_(False)
        )
        self.normalize = transforms.Normalize(
            mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225)
        )
        self.size = 224, 224

    @classmethod
    def available_models(cls):
        return ["vits14", "vitb14", "vitl14", "vitg14"]

    def forward(self, x):
        if x.shape[1] == 1:
            x = torch.cat([x] * 3, dim=1)
        x = self.normalize(x)
        with torch.autocast("cuda", dtype=torch.float16):
            x = self.model(x.cuda()).float()
        # With * x.shape[-1] ** 0.5, the result will be much larger and related to the model size(emb dim)
        # in here we only use normalize to get comparable result across different model sizes
        x = F.normalize(x)  # * x.shape[-1] ** 0.5
        return x


class FIDRunner(MetricRunner):
    multi = True

    def __init__(self, feature=2048, img_size=(299, 299)):
        self.model = (
            FrechetInceptionDistance(
                feature=feature,
                reset_real_features=False,
                normalize=True,
                input_img_size=(3, *img_size),
            )
            .eval()
            .requires_grad_(False)
            .cuda()
        )
        self.img_size = img_size
        self.img_load_func = partial(load_img, size=self.img_size)

    @torch.no_grad()
    def eval(self, images, ref_texts=None, is_ref=False):
        images = torch.stack(images).cuda()
        self.model.update(images, is_ref)

    @torch.no_grad()
    def eval_multi(self, images, ref_texts=None, ref_images=None, batch_size=32):
        super().eval_multi(images, ref_texts, ref_images, batch_size)
        result = self.model.compute()
        self.model.reset()
        return result


if __name__ == "__main__":
    model = sys.argv[1] if len(sys.argv) > 1 else "vitb14"
    swinv2 = DINOv2FeatureExtractor(model)
    runner = FIDRunner(swinv2, (224, 224))
    print(model)

    PATH = r"F:\dataset\HakuBooru\out\scenery"
    img_files = [os.path.join(PATH, i) for i in os.listdir(PATH)]
    refs = [i for i in img_files if i.endswith(".webp")]  # [:32768]
    print("ref image count:", len(refs))

    PATH = "./data/scenery-tag"
    results = {}
    for idx, folder in enumerate([i for i in os.listdir(PATH) if i =="Prompt-DB"]):
        img_files = [
            os.path.join(PATH, folder, i)
            for i in os.listdir(os.path.join(PATH, folder))
        ]

        images = [i for i in img_files if i.endswith(".webp")]
        print(folder, len(images))
        result = runner.eval_multi(
            images, ref_images=refs if not idx else [], batch_size=512
        )
        results[folder] = result

    print("=" * 20)
    for folder, result in results.items():
        print(f"{folder:<10}:", result.item())



from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from PIL import Image
from tqdm import tqdm, trange


pool = ProcessPoolExecutor(max_workers=24)
# pool = ThreadPoolExecutor(max_workers=64)


def load(image):
    if isinstance(image, str):
        return Image.open(image).convert("RGB")
    else:
        return image


def batch_load(images, loading_func=load):
    return list(tqdm(pool.map(loading_func, images), total=len(images), leave=False))


class MetricRunner:
    single = False
    multi = False
    img_load_func = load

    def eval(self, images, ref_texts=None, is_ref=False):
        raise NotImplementedError()

    def eval_single(self, image, ref_text=None, is_ref=False):
        if isinstance(image, (Image.Image, str)):
            image = [image]
        if isinstance(ref_text, str):
            ref_text = [ref_text]
        image = batch_load(image, self.img_load_func)

        return self.eval(image, ref_text, is_ref=is_ref)

    def eval_multi(self, images, ref_texts=None, ref_images=None, batch_size=32):
        if ref_texts is None:
            ref_texts = [None] * max(
                len(images), 0 if ref_images is None else len(ref_images)
            )
        results = []
        ref_results = []
        for i in trange(0, len(images), batch_size):
            results.append(
                self.eval_single(
                    images[i : i + batch_size],
                    ref_texts[i : i + batch_size],
                    is_ref=False,
                )
            )
        if ref_images is not None:
            for i in trange(0, len(ref_images), batch_size):
                ref_results.append(
                    self.eval_single(
                        ref_images[i : i + batch_size],
                        ref_texts[i : i + batch_size],
                        is_ref=True,
                    )
                )

        return results, ref_results



import random
from typing import Iterable


def shuffle_iterable(x: Iterable) -> list:
    x_list = list(x)
    random.shuffle(x_list)
    return x_list


def same_order_deduplicate(x: list):
    new_list = []
    history = set()
    for i in x:
        if i not in history:
            new_list.append(i)
            history.add(i)
    return new_list


def remove_repeated_suffix(s):
    """
    Removes the repeated suffix from the string efficiently using Rolling Hash.

    Args:
        s (str): The input string.

    Returns:
        str: The string with the repeated suffix removed.
    """
    if not s:
        return s

    num_chars = len(s)
    base = 257  # A prime number base for hashing
    modulus = 10**9 + 7  # A large prime modulus to prevent overflow

    # Precompute prefix hashes and powers of the base
    prefix_hashes = [0] * (num_chars + 1)
    powers = [1] * (num_chars + 1)

    for i in range(num_chars):
        prefix_hashes[i + 1] = (prefix_hashes[i] * base + ord(s[i])) % modulus
        powers[i + 1] = (powers[i] * base) % modulus

    def get_hash(start, end):
        return (
            prefix_hashes[end] - prefix_hashes[start] * powers[end - start]
        ) % modulus

    max_suffix_length = 0  # To store the maximum k where suffix is repeated

    # Iterate over possible suffix lengths from 1 to num_chars//2
    for suffix_length in range(1, num_chars // 2 + 1):
        # Compare the last suffix_length characters with the suffix_length characters before them
        if get_hash(
            num_chars - 2 * suffix_length, num_chars - suffix_length
        ) == get_hash(num_chars - suffix_length, num_chars):
            max_suffix_length = (
                suffix_length  # Update max_suffix_length if a repeated suffix is found
            )

    if max_suffix_length > 0:
        # Remove the extra occurrences of the suffix
        # Calculate how many times the suffix is repeated consecutively
        num_repetitions = 2
        while max_suffix_length * (num_repetitions + 1) <= num_chars and get_hash(
            num_chars - (num_repetitions + 1) * max_suffix_length,
            num_chars - num_repetitions * max_suffix_length,
        ) == get_hash(
            num_chars - num_repetitions * max_suffix_length,
            num_chars - (num_repetitions - 1) * max_suffix_length,
        ):
            num_repetitions += 1
        # Remove (num_repetitions-1) copies of the suffix
        s = s[: num_chars - (num_repetitions - 1) * max_suffix_length]

    return s


if __name__ == "__main__":
    text = "123ababababab"
    print(remove_repeated_suffix(text))



import os
import numpy
import orjsonl
import orjson
from PIL import Image
from sdeval.corrupt import AICorruptMetrics
from sdeval.corrupt.aicorrupt import _DEFAULT_MODEL_NAME

from .base import MetricRunner, load


class AICorruptRunner(MetricRunner):
    single = True
    multi = True

    def __init__(self, model_name=_DEFAULT_MODEL_NAME):
        # load model and preprocessor
        self.model = AICorruptMetrics(model_name=model_name, silent=True)

    def img_load_func(self, image):
        image = load(image)
        image = image.resize((384, 384), Image.BICUBIC)
        return image

    def eval(self, images, ref_texts=None, is_ref=False):
        return self.model.score(images, mode="seq")

    def eval_multi(self, images, ref_texts=None, ref_images=None, batch_size=32):
        results, ref_results = super().eval_multi(images, ref_texts, ref_images, batch_size)
        results = numpy.concatenate(results, axis=0)
        return [float(i) for i in results]


if __name__ == "__main__":
    runner = AICorruptRunner()
    images = [None] * 4

    PATH = "./data/coyo-img"
    img_files = [os.path.join(PATH, i) for i in os.listdir(PATH)]
    images[0] = [i for i in img_files if i.endswith("_short.webp")]
    images[1] = [i for i in img_files if i.endswith("_truncate_long.webp")]
    images[2] = [i for i in img_files if i.endswith("_gen.webp")]
    images[3] = [i for i in img_files if i.endswith("_extend.webp")]

    with open("coyo-ai-corrupt.jsonl", "w") as f:
        for idx, name in enumerate(["short", "truncate", "tipo gen", "tipo extend"]):
            result = runner.eval_multi(images[idx], batch_size=500)
            f.write(str(result) + "\n")



DEFAULT_NEGATIVE_PROMPT = """
low quality, worst quality, normal quality, text, signature, jpeg artifacts, 
bad anatomy, old, early, mini skirt, nsfw, chibi, multiple girls, multiple boys, 
multiple tails, multiple views, copyright name, watermark, artist name, signature
"""

DEFAULT_FORMAT = {
    "tag only (DTG mode)": """
<|special|>, <|characters|>, <|copyrights|>, 
<|artist|>, 

<|general|>,

<|quality|>, <|meta|>, <|rating|>
""".strip(),
    "NL only (Tag to NL)": """<|extended|>.""".strip(),
    "Both, tag first (recommend)": """
<|special|>, <|characters|>, <|copyrights|>, 
<|artist|>, 

<|general|>,

<|extended|>.

<|quality|>, <|meta|>, <|rating|>
""".strip(),
    "Both, NL first (recommend)": """
<|special|>, <|characters|>, <|copyrights|>, 
<|artist|>, 

<|extended|>.

<|general|>,

<|quality|>, <|meta|>, <|rating|>
""".strip(),
    "Both + generated NL": """
<|special|>, <|characters|>, <|copyrights|>, 
<|artist|>, 

<|generated|>.

<|general|>,

<|extended|>.

<|quality|>, <|meta|>, <|rating|>
""".strip(),
}



import math
from functools import partial
from time import time
from PIL import Image

import numpy as np
import torch
from diffusers import (
    StableDiffusionXLKDiffusionPipeline,
    UNet2DConditionModel,
    AutoencoderKL,
)
from k_diffusion.external import CompVisDenoiser
from k_diffusion.sampling import get_sigmas_polyexponential
from k_diffusion.sampling import sample_dpmpp_2m_sde

torch.set_float32_matmul_precision("medium")


def set_timesteps_polyexponential(
    self, orig_sigmas, num_inference_steps, device=None, rho=None
):
    self.num_inference_steps = num_inference_steps

    self.sigmas = get_sigmas_polyexponential(
        num_inference_steps + 1,
        sigma_min=orig_sigmas[-2],
        sigma_max=orig_sigmas[0],
        rho=rho or 0.666666,
        device=device or "cpu",
    )
    self.sigmas = torch.cat([self.sigmas[:-2], self.sigmas.new_zeros([1])])


def set_timesteps_exponential(self, orig_sigmas, num_inference_steps, device=None):
    return set_timesteps_polyexponential(
        self, orig_sigmas, num_inference_steps, device, 1.0
    )


def set_timesteps_linear(self, orig_sigmas, num_inference_steps, device=None):
    self.num_inference_steps = num_inference_steps
    index = (
        torch.linspace(0, orig_sigmas.numel() - 1, num_inference_steps).round().long()
    )
    self.sigmas = torch.cat([orig_sigmas[index], self.sigmas.new_zeros([1])])


def model_forward(k_diffusion_model: torch.nn.Module):
    orig_forward = k_diffusion_model.forward

    def forward(*args, **kwargs):
        with torch.autocast(device_type="cuda", dtype=torch.float16):
            result = orig_forward(*args, **kwargs)
        return result.float()

    return forward


def load_model(model_id="KBlueLeaf/Kohaku-XL-Zeta", device="cuda", custom_vae=False):
    if custom_vae:
        vae: AutoencoderKL = AutoencoderKL.from_pretrained(
            "madebyollin/sdxl-vae-fp16-fix"
        ).to(device)
        pipe: StableDiffusionXLKDiffusionPipeline
        pipe = StableDiffusionXLKDiffusionPipeline.from_pretrained(
            model_id, torch_dtype=torch.float16, vae=vae
        ).to(device)
    else:
        pipe: StableDiffusionXLKDiffusionPipeline
        pipe = StableDiffusionXLKDiffusionPipeline.from_pretrained(
            model_id, torch_dtype=torch.float16
        ).to(device)
    pipe.vae.config.force_upcast = False
    pipe.vae.eval().half()
    pipe.text_encoder.eval().half()
    pipe.text_encoder_2.eval().half()
    pipe.k_diffusion_model.eval().half().to(device)
    unet: UNet2DConditionModel = pipe.k_diffusion_model.inner_model.model
    unet.eval().half()
    unet.enable_xformers_memory_efficient_attention()
    pipe.scheduler.set_timesteps = partial(
        set_timesteps_exponential, pipe.scheduler, pipe.scheduler.sigmas
    )
    # pipe.scheduler.set_timesteps = partial(
    #     set_timesteps_linear, pipe.scheduler, pipe.scheduler.sigmas
    # )
    pipe.sampler = partial(sample_dpmpp_2m_sde, eta=0.35, solver_type="heun")
    pipe.k_diffusion_model.forward = model_forward(pipe.k_diffusion_model)
    return pipe


@torch.no_grad()
def encode_prompts(
    pipe: StableDiffusionXLKDiffusionPipeline,
    prompt: str | list[str],
    neg_prompt: str | list[str] = "",
    cutoff_length: int | None = 225,
    padding_to_max_length: bool = True,
    take_all_eos: bool = False,
):
    if not isinstance(prompt, list):
        prompt = [prompt]
    if not isinstance(neg_prompt, list):
        neg_prompt = [neg_prompt]
    if len(prompt) != len(neg_prompt) and (len(prompt) != 1 and len(neg_prompt) != 1):
        raise ValueError("prompt and neg_prompt must have the same length")
    if len(prompt) == 1:
        prompt = prompt * len(neg_prompt)
    if len(neg_prompt) == 1:
        neg_prompt = neg_prompt * len(prompt)
    prompts = prompt + neg_prompt
    max_length = pipe.tokenizer.model_max_length - 2

    input_ids = pipe.tokenizer(prompts, padding=True, return_tensors="pt")
    input_ids2 = pipe.tokenizer_2(prompts, padding=True, return_tensors="pt")
    input_ids_neg = pipe.tokenizer(neg_prompt, padding=True, return_tensors="pt")
    length = max(input_ids.input_ids.size(-1), input_ids2.input_ids.size(-1))
    neg_length = input_ids_neg.input_ids.size(-1)
    target_length = cutoff_length or math.ceil(length / max_length) * max_length + 2
    neg_groups = math.ceil(neg_length / max_length)

    input_ids = pipe.tokenizer(
        prompts,
        padding="max_length",
        max_length=target_length,
        truncation=True,
        return_tensors="pt",
    ).input_ids
    input_ids = (
        input_ids[:, 0:1],
        input_ids[:, 1:-1],
        input_ids[:, -1:],
    )
    input_ids2 = pipe.tokenizer_2(
        prompts,
        padding="max_length",
        max_length=target_length,
        truncation=True,
        return_tensors="pt",
    ).input_ids
    input_ids2 = (
        input_ids2[:, 0:1],
        input_ids2[:, 1:-1],
        input_ids2[:, -1:],
    )

    concat_embeds = []
    for i in range(0, input_ids[1].shape[-1], max_length):
        input_id1 = torch.concat(
            (input_ids[0], input_ids[1][:, i : i + max_length], input_ids[2]), dim=-1
        ).to(pipe.device)
        result = pipe.text_encoder(input_id1, output_hidden_states=True).hidden_states[
            -2
        ]
        if take_all_eos:
            concat_embeds.append(result)
            continue
        if i == 0:
            concat_embeds.append(result[:, :-1])
        elif i == input_ids[1].shape[-1] - max_length:
            concat_embeds.append(result[:, 1:])
        else:
            concat_embeds.append(result[:, 1:-1])

    concat_embeds2 = []
    pooled_embeds2 = []
    for i in range(0, input_ids2[1].shape[-1], max_length):
        input_id2 = torch.concat(
            (input_ids2[0], input_ids2[1][:, i : i + max_length], input_ids2[2]), dim=-1
        ).to(pipe.device)
        hidden_states = pipe.text_encoder_2(input_id2, output_hidden_states=True)
        pooled_embeds2.append(hidden_states[0])
        if take_all_eos:
            concat_embeds2.append(hidden_states.hidden_states[-2])
            continue
        if i == 0:
            concat_embeds2.append(hidden_states.hidden_states[-2][:, :-1])
        elif i == input_ids2[1].shape[-1] - max_length:
            concat_embeds2.append(hidden_states.hidden_states[-2][:, 1:])
        else:
            concat_embeds2.append(hidden_states.hidden_states[-2][:, 1:-1])

    prompt_embeds = torch.cat(concat_embeds, dim=1)
    prompt_embeds2 = torch.cat(concat_embeds2, dim=1)
    prompt_embeds = torch.cat([prompt_embeds, prompt_embeds2], dim=-1)

    pooled = torch.mean(torch.stack(pooled_embeds2, dim=0), dim=0)

    embed, neg_embed = prompt_embeds.chunk(2)
    pooled, neg_pooled = pooled.chunk(2)
    if not padding_to_max_length:
        if take_all_eos:
            neg_embed = torch.cat([concat_embeds[0][1:], concat_embeds2[0][1:]], dim=-1)
        else:
            neg_embed = torch.cat(
                [neg_embed[:, : neg_groups * max_length + 1, :], neg_embed[:, -1:, :]],
                dim=1,
            )
        neg_pooled_embeds = torch.stack(
            [emb.chunk(2)[-1] for emb in pooled_embeds2][:neg_groups]
        )
        neg_pooled = torch.mean(neg_pooled_embeds, dim=0)

    return (embed, neg_embed), (pooled, neg_pooled)


def vae_image_postprocess(image_tensor: torch.Tensor) -> Image.Image:
    image = Image.fromarray(
        ((image_tensor * 0.5 + 0.5) * 255)
        .cpu()
        .clamp(0, 255)
        .numpy()
        .astype(np.uint8)
        .transpose(1, 2, 0)
    )
    return image


@torch.no_grad()
def generate(
    pipe: StableDiffusionXLKDiffusionPipeline,
    prompt_embeds,
    negative_prompt_embeds,
    pooled_prompt_embeds,
    negative_pooled_prompt_embeds,
    num_inference_steps=24,
    width=1024,
    height=1024,
    guidance_scale=6.0,
):
    pipe.scheduler.set_timesteps(num_inference_steps)
    unet: CompVisDenoiser = pipe.k_diffusion_model

    if prompt_embeds.shape == negative_prompt_embeds.shape:
        time_ids = (
            torch.tensor([height, width, 0, 0, height, width])
            .repeat(2 * prompt_embeds.size(0), 1)
            .to(prompt_embeds)
        )
        added_cond = {
            "time_ids": time_ids,
            "text_embeds": torch.concat(
                [pooled_prompt_embeds, negative_pooled_prompt_embeds]
            ),
        }
        text_ctx = torch.cat([prompt_embeds, negative_prompt_embeds])

        def cfg_wrapper(x, sigma):
            cond, uncond = unet(
                torch.cat([x] * 2),
                torch.cat([sigma] * 2),
                cond=text_ctx,
                added_cond_kwargs=added_cond,
            ).chunk(2)
            cfg_output = uncond + guidance_scale * (cond - uncond)
            return cfg_output

    else:
        time_ids = torch.tensor([height, width, 0, 0, height, width]).to(prompt_embeds)
        added_cond_pos = {"time_ids": time_ids, "text_embeds": pooled_prompt_embeds}
        added_cond_neg = {
            "time_ids": time_ids,
            "text_embeds": negative_pooled_prompt_embeds,
        }

        def cfg_wrapper(x, sigma):
            cond = unet(x, sigma, cond=prompt_embeds, added_cond_kwargs=added_cond_pos)
            uncond = unet(
                x, sigma, cond=negative_prompt_embeds, added_cond_kwargs=added_cond_neg
            )
            cfg_output = uncond + guidance_scale * (cond - uncond)
            return cfg_output

    sigmas = pipe.scheduler.sigmas
    x0 = (
        torch.randn(
            (1, 4, height // 8, width // 8),
        )
        .to(prompt_embeds)
        .repeat(prompt_embeds.size(0), 1, 1, 1)
        * sigmas[0]
    )
    result = sample_dpmpp_2m_sde(
        cfg_wrapper, x0, sigmas.to(prompt_embeds.device), eta=1.0
    )
    result /= pipe.vae.config.scaling_factor
    image_tensors = []
    for latent in result:
        image_tensors.append(pipe.vae.decode(latent.unsqueeze(0).half()).sample)
    image_tensors = torch.concat(image_tensors)
    images = []
    for image_tensor in image_tensors:
        images.append(vae_image_postprocess(image_tensor))
    return images


if __name__ == "__main__":
    prompt = """
1girl,
king halo (umamusume), umamusume,

ogipote, misu kasumi, fuzichoco, ningen mame, ask (askzy), maccha (mochancc),

solo, leaning forward, cleavage, sky, outdoors, black bikini, stomach, swimsuit, navel, 

masterpiece, newest, absurdres, sensitive
""".strip()
    # sdxl_pipe = load_model("KBlueLeaf/xxx")
    sdxl_pipe = load_model()
    t0 = time()
    for _ in range(10):
        with torch.autocast("cuda"):
            (prompt_embeds, neg_prompt_embeds), (pooled_embeds2, neg_pooled_embeds2) = (
                encode_prompts(sdxl_pipe, [prompt] * 3, "")
            )
            result = generate(
                sdxl_pipe,
                prompt_embeds,
                neg_prompt_embeds,
                pooled_embeds2,
                neg_pooled_embeds2,
                num_inference_steps=24,
                width=1024,
                height=1024,
                guidance_scale=6.0,
            )[0]
    t1 = time()
    print((t1 - t0) / 100)

    result.save("output/test.png")
